/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * PrimeBase Media Stream for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Original author: Paul McCullagh
 * Continued development: Barry Leslie
 *
 * 2007-05-25
 *
 * H&G2JCtL
 *
 * Network interface.
 *
 */

#include "CSConfig.h"

#include <string.h>
#include <ctype.h>

#include "string.h"

#include "CSGlobal.h"
#include "CSLog.h"
#include "CSDirectory.h"
#include "CSStrUtil.h"

#include "Database_ms.h"
#include "OpenTable_ms.h"
#include "backup_ms.h"
#include "Table_ms.h"
#include "Util_ms.h"
#include "TempLog_ms.h"
#include "Network_ms.h"
#include "ms_mysql.h"
#include "pbmslib.h"
#include "Transaction_ms.h"

csWord8				MSDatabase::gRepoThreshold;
csWord8				MSDatabase::gTempLogThreshold;
CSSyncStringBuffer	*MSDatabase::gDefaultMetaDataHeaders;
CSSyncSortedList	*MSDatabase::gDatabaseList;
CSSparseArray		*MSDatabase::gDatabaseArray;
csWord8				MSDatabase::gBackupDatabaseID;
/*
 * -------------------------------------------------------------------------
 * PBMS DATABASES
 */

MSDatabase::MSDatabase():
myDatabasePath(NULL),
myDatabaseName(NULL),
myTempLogArray(NULL),
myCompactorThread(NULL),
myTempLogThread(NULL),
myRepostoryList(NULL),
myBlobCloud(NULL),
iTableList(NULL),
iTableArray(NULL),
iMaxTableID(0),
iBlobAliases(NULL),
iDropping(false),
iRecovering(false),
isBackup(false),
iWriteTempLog(NULL),
iBackupThread(NULL),
iNextBlobRefId(0),
myBackupLocation(NULL),
iClosing(false)
{
}

MSDatabase::~MSDatabase()
{
	iClosing = true;
	if (iBackupThread) {
		iBackupThread->stop();
		iBackupThread->release();
		iBackupThread = NULL;
	}
	
	if (myTempLogThread) {
		myTempLogThread->stop();
		myTempLogThread->release();
		myTempLogThread = NULL;
	}
	if (myCompactorThread) {
		myRepostoryList->wakeup(); // The compator thread waits on this.
		myCompactorThread->stop();
		myCompactorThread->release();
		myCompactorThread = NULL;
	}
	
	if (myDatabasePath)
		myDatabasePath->release();
		
	if (myDatabaseName)
		myDatabaseName->release();
		
	iWriteTempLog = NULL;
	if (myTempLogArray) {
		myTempLogArray->clear();
		myTempLogArray->release();
	}
	if (iTableList) {
		iTableList->clear();
		iTableList->release();
	}
	if (iTableArray){
		iTableArray->clear();
		iTableArray->release();
	}
	if (myRepostoryList) {
		myRepostoryList->clear();
		myRepostoryList->release();
	}
	if (iBlobAliases) {
		iBlobAliases->ma_close();
		iBlobAliases->release();
	}
	if (myBlobCloud) {
		myBlobCloud->release();
	}
	if (myBackupLocation) {
		myBackupLocation->release();
	}
}

const char *MSDatabase::getDatabaseNameCString()
{
	return myDatabaseName->getCString();
}

MSTable *MSDatabase::getTable(CSString *tab_name, bool create)
{
	MSTable *tab;
	
	enter_();
	push_(tab_name);
	lock_(iTableList);
	if (!(tab = (MSTable *) iTableList->find(tab_name))) {
		/* Check if such a table exists:
		 * I have to do this, because after rename, a request
		 * my come immediately, before I have a chance to
		 * open it!
		 *
		 * --- Commented out.
		 * By doing an atomic rename in the database,
		 * this should not longer be necessary.
		CSDirectory		*dir;
		char			dir_tab_name[MS_TABLE_NAME_SIZE];
		const char		*file_name;

		myDatabasePath->retain();
		dir = CSDirectory::newDirectory(myDatabasePath);
		push_(dir);
		dir->open();
		while (dir->next()) {
			file_name = dir->name();
			if (dir->isFile() && cs_is_extension(file_name, "bst")) {
				ms_file_to_table_name(MS_TABLE_NAME_SIZE, dir_tab_name, file_name);
				if (tab_name->compare(dir_tab_name) == 0)
					// Found the table:
					addTableFromFile(dir, file_name, false);
			}
		}
		release_(dir);
		if (!tab && create) {
		 */
		if (create) {
			/* Create a new table: */
			tab_name->retain();
			tab = MSTable::newTable(iMaxTableID+1, tab_name, this, (off_t) 0, false);
			iTableList->add(tab);
			tab->retain();
			iTableArray->set(iMaxTableID+1, tab);
			iMaxTableID++;
		}
	}
	if (tab)
		tab->retain();
	unlock_(iTableList);
	release_(tab_name);
	return_(tab);
}

MSTable *MSDatabase::getTable(const char *tab_name, bool create)
{
	return getTable(CSString::newString(tab_name), create);
}


MSTable *MSDatabase::getTable(csWord4 tab_id, bool missing_ok)
{
	MSTable *tab;
	
	enter_();
	lock_(iTableList);
	if (!(tab = (MSTable *) iTableArray->get((u_int) tab_id))) {
		if (missing_ok) {
			unlock_(iTableList);
			return_(NULL);
		}
		char buffer[CS_EXC_MESSAGE_SIZE];

		cs_strcpy(CS_EXC_MESSAGE_SIZE, buffer, "Unknown table #");
		cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, (u_int) tab_id);
		cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, " in database ");
		cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, getDatabaseNameCString());
		CSException::throwException(CS_CONTEXT, MS_ERR_UNKNOWN_TABLE, buffer);
	}
	tab->retain();
	unlock_(iTableList);
	return_(tab);
}

MSTable *MSDatabase::getNextTable(csWord4 *pos)
{
	u_int i = *pos;
	MSTable *tab = NULL;
	
	enter_();
	lock_(iTableList);
	while (i < iTableList->getSize()) {
		tab = (MSTable *) iTableList->itemAt(i++);
		if (!tab->isToDelete())
			break;
		tab = NULL;
	}
	if (tab)
		tab->retain();
	unlock_(iTableList);
	*pos = i;
	return_(tab);
}

void MSDatabase::addTable(csWord4 tab_id, const char *tab_name, off_t file_size, bool to_delete)
{
	MSTable	*tab;

	if (tab_id > iMaxTableID)
		iMaxTableID = tab_id;
	tab = MSTable::newTable(tab_id, tab_name, this, file_size, to_delete);
	iTableList->add(tab);
	tab->retain();
	iTableArray->set(tab_id, tab);
}

void MSDatabase::addTableFromFile(CSDirectory *dir, const char *file_name, bool to_delete)
{
	off_t	file_size;
	u_int	file_id;
	char	tab_name[MS_TABLE_NAME_SIZE];

	dir->info(NULL, &file_size, NULL);
	file_id = ms_file_to_table_id(file_name);
	ms_file_to_table_name(MS_TABLE_NAME_SIZE, tab_name, file_name);
	addTable(file_id, tab_name, file_size, to_delete);
}

void MSDatabase::removeTable(MSTable *tab)
{
	enter_();
	push_(tab);
	lock_(iTableList);
	iTableList->remove(tab->myTableName);
	iTableArray->remove(tab->myTableID);
	unlock_(iTableList);
	release_(tab);
	exit_();
}

void MSDatabase::dropTable(MSTable *tab)
{
	enter_();
	push_(tab);
	lock_(iTableList);
	iTableList->remove(tab->myTableName);
	iTableArray->remove(tab->myTableID);

	// Cute: you drop the table by adding it with the 'to_delete' flag set to 'true'
	addTable(tab->myTableID, tab->myTableName->getCString(), tab->getTableFileSize(), true);

	unlock_(iTableList);
	release_(tab);
	exit_();
}

// This function is used when dropping tables from a database before
// dropping the database itself. 
CSString *MSDatabase::getATableName()
{
	u_int i = 0;
	MSTable *tab;
	CSString *name = NULL;
	
	enter_();
	lock_(iTableList);

	while ((tab = (MSTable *) iTableList->itemAt(i++)) && tab->isToDelete()) ;
	if (tab) {
		name = tab->getTableName();
		name->retain();
	}
	unlock_(iTableList);
	return_(name);
}

u_int MSDatabase::getTableCount()
{
	u_int cnt = 0, i = 0;
	MSTable *tab;
	
	enter_();
	lock_(iTableList);

	while ((tab = (MSTable *) iTableList->itemAt(i++))) {
		if (!tab->isToDelete())
			cnt++;
	}

	unlock_(iTableList);
	return_(cnt);
}


void MSDatabase::renameTable(MSTable *tab, const char *to_name)
{
	enter_();
	lock_(iTableList);
	iTableList->remove(tab->myTableName);
	iTableArray->remove(tab->myTableID);

	addTable(tab->myTableID, to_name, tab->getTableFileSize(), false);

	unlock_(iTableList);
	exit_();
}

void MSDatabase::openWriteRepo(MSOpenTable *otab)
{
	if (otab->myWriteRepo && otab->myWriteRepoFile)
		return;

	enter_();
	if (!otab->myWriteRepo)
		otab->myWriteRepo = lockRepo(0);

	/* Now open the repo file for the open table: */
	otab->myWriteRepo->openRepoFileForWriting(otab);
	exit_();
}

MSRepository *MSDatabase::getRepoFullOfTrash(time_t *ret_wait_time)
{
	MSRepository	*repo = NULL;
	time_t			wait_time = 0;
	
	if (ret_wait_time)
		wait_time = *ret_wait_time;
	enter_();
	lock_(myRepostoryList);
	for (u_int i=0; i<myRepostoryList->size(); i++) {
		retry:
		if ((repo = (MSRepository *) myRepostoryList->get(i))) {
			if (!repo->isRemovingFP && !repo->mustBeDeleted && !repo->isRepoLocked()) {
				if (!repo->myRepoHeadSize) {
					/* The file has not yet been opened, so the
					 * garbage count will not be known!
					 */
					MSRepoFile *repo_file;

					repo->retain();
					unlock_(myRepostoryList);
					push_(repo);
					repo_file = repo->openRepoFile();
					repo_file->release();
					release_(repo);
					lock_(myRepostoryList);
					goto retry;
				}
				if (repo->getGarbageLevel() >= MSRepository::gGarbageThreshold) {
					/* Make sure there are not temp BLOBs in this repository that have
					 * not yet timed out:
					 */
					time_t now = time(NULL);
					time_t then = repo->myLastTempTime;

					/* Check if there are any temp BLOBs to be removed: */
					if (now > then + MSTempLog::gTempBlobTimeout) {
						repo->lockRepo(REPO_COMPACTING); 
						repo->retain();
						break;
					}
					else {
						/* There are temp BLOBs to wait for... */
						if (!wait_time || wait_time > MSTempLog::adjustWaitTime(then, now))
							wait_time = MSTempLog::adjustWaitTime(then, now);
					}
				}
			}
			repo = NULL;
		}
	}
	unlock_(myRepostoryList);
	if (ret_wait_time)
		*ret_wait_time = wait_time;
	return_(repo);
}

MSRepository *MSDatabase::lockRepo(off_t size)
{
	MSRepository	*repo;
	u_int			free_slot;

	enter_();
	lock_(myRepostoryList);
	free_slot = myRepostoryList->size();
	/* Find an unlocked repository file that is below the write threshold: */
	for (u_int i=0; i<myRepostoryList->size(); i++) {
		if ((repo = (MSRepository *) myRepostoryList->get(i))) {
			if (!repo->isRepoLocked() && !repo->isRemovingFP && !repo->mustBeDeleted &&
				repo->myRepoFileSize + size < gRepoThreshold 
				/**/ && repo->getGarbageLevel() < MSRepository::gGarbageThreshold)
				goto found1;
		}
		else {
			if (i < free_slot)
				free_slot = i;
		}
	}

	/* None found, create a new repo file: */
	new_(repo, MSRepository(free_slot + 1, this, 0));
	myRepostoryList->set(free_slot, repo);

	found1:
	repo->retain();
	repo->lockRepo(REPO_WRITE); 
	unlock_(myRepostoryList);
	return_(repo);
}

MSRepoFile *MSDatabase::getRepoFileFromPool(csWord4 repo_id, bool missing_ok)
{
	MSRepository	*repo;
	MSRepoFile		*file;

	enter_();
	lock_(myRepostoryList);
	if (!(repo = (MSRepository *) myRepostoryList->get(repo_id - 1))) {
		if (!missing_ok) {
			char buffer[CS_EXC_MESSAGE_SIZE];

			cs_strcpy(CS_EXC_MESSAGE_SIZE, buffer, "Unknown repository file: ");
			cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, (u_int) repo_id);
			CSException::throwException(CS_CONTEXT, MS_ERR_NOT_FOUND, buffer);
		}
		unlock_(myRepostoryList);
		return_(NULL);
	}
	if (repo->isRemovingFP) {
		char buffer[CS_EXC_MESSAGE_SIZE];

		cs_strcpy(CS_EXC_MESSAGE_SIZE, buffer, "Repository will be removed: ");
		cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, (u_int) repo_id);
		CSException::throwException(CS_CONTEXT, MS_ERR_REMOVING_REPO, buffer);
	}
	repo->retain(); /* Release is here: [++] */
	file = repo->getRepoFile();
	unlock_(myRepostoryList);

	if (!file) {
		file = repo->openRepoFile();
		lock_(myRepostoryList);
		repo->addRepoFile(file);
		file->retain();
		unlock_(myRepostoryList);
	}
	return_(file);
}

void MSDatabase::returnRepoFileToPool(MSRepoFile *file)
{
	MSRepository	*repo;

	enter_();
	lock_(myRepostoryList);
	push_(file);
	if ((repo = file->myRepo)) {
		if (repo->isRemovingFP) {
			repo->removeRepoFile(file);
			myRepostoryList->wakeup();
		}
		else
			repo->returnRepoFile(file);
		repo->release(); /* [++] here is the release.  */
	}
	release_(file);
	unlock_(myRepostoryList);
	exit_();
}

void MSDatabase::removeRepo(csWord4 repo_id, bool *mustQuit)
{
	MSRepository *repo;

	enter_();
	lock_(myRepostoryList);
	while ((!mustQuit || !*mustQuit) && !iClosing) {
		if (!(repo = (MSRepository *) myRepostoryList->get(repo_id - 1)))
			break;
		repo->isRemovingFP = true;
		if (repo->removeRepoFilesNotInUse()) {
			myRepostoryList->set(repo_id - 1, NULL);
			break;
		}
		/*
		 * Wait for the files that are in use to be
		 * freed.
		 */
		myRepostoryList->wait();
	}
	unlock_(myRepostoryList);
	exit_();
}

void MSDatabase::queueTempLogEvent(MSOpenTable *otab, int type, csWord4 tab_id, csWord8 blob_id, csWord4 auth_code,
	 csWord4 *log_id, csWord4 *log_offset, csWord4 *q_time)
{
	MSTempLogItemRec	item;
	csWord4				timev;

	// Each otab object holds an handle to an instance of an OPEN
	// temp log. This is so that each thread has it's own open temp log 
	// and doesn't need to be opened and close it constantly.
	
	enter_();
	lock_(myTempLogArray);
	if (!iWriteTempLog) {
		iWriteTempLog = (MSTempLog *) myTempLogArray->last();
		if (!iWriteTempLog) {
			new_(iWriteTempLog, MSTempLog(1, this, 0));
			myTempLogArray->set(1, iWriteTempLog);
		}
	}
	if (!otab->myTempLogFile)
		otab->myTempLogFile = iWriteTempLog->openTempLog();
	else if (otab->myTempLogFile->myTempLogID != iWriteTempLog->myLogID) {
		otab->myTempLogFile->release();
		otab->myTempLogFile = NULL;
		otab->myTempLogFile = iWriteTempLog->openTempLog();
	}

	if (iWriteTempLog->myTempLogSize >= gTempLogThreshold) {
		u_int log_id = iWriteTempLog->myLogID + 1;

		new_(iWriteTempLog, MSTempLog(log_id, this, 0));
		myTempLogArray->set(log_id, iWriteTempLog);

		otab->myTempLogFile->release();
		otab->myTempLogFile = NULL;
		otab->myTempLogFile = iWriteTempLog->openTempLog();

	}

	timev = time(NULL);
	*log_id = iWriteTempLog->myLogID;
	*log_offset = (csWord4) iWriteTempLog->myTempLogSize;
	if (q_time)
		*q_time = timev;
	iWriteTempLog->myTempLogSize += iWriteTempLog->myTemplogRecSize;
	unlock_(myTempLogArray);

	CS_SET_DISK_1(item.ti_type_1, type);
	CS_SET_DISK_4(item.ti_table_id_4, tab_id);
	CS_SET_DISK_6(item.ti_blob_id_6, blob_id);
	CS_SET_DISK_4(item.ti_auth_code_4, auth_code);
	CS_SET_DISK_4(item.ti_time_4, timev);
	otab->myTempLogFile->write(&item, *log_offset, sizeof(MSTempLogItemRec));
	
	
	exit_();
}

void MSDatabase::queueForDeletion(MSOpenTable *otab, int type, csWord4 tab_id, csWord8 blob_id, csWord4 auth_code,
	 csWord4 *log_id, csWord4 *log_offset, csWord4 *q_time, MSDiskAliasPtr aliasDiskRec)
{
	enter_();
	
	queueTempLogEvent(otab, type, tab_id, blob_id, auth_code, log_id, log_offset, q_time);
		
	// If it has an alias remove it from the ailias index.
	if (aliasDiskRec) {
		try_(a) {
			deleteBlobAlias(aliasDiskRec);
		}
		catch_(a);
		self->logException();
		cont_(a);
	}
	
	exit_();
}

void MSDatabase::queueCloudEvent(MSOpenTable *otab, csWord4 tab_id, csWord8 blob_id, csWord4 auth_code,
	 csWord4 *log_id, csWord4 *log_offset, csWord4 *q_time)
{
	enter_();

	if (myBlobCloud)
		queueTempLogEvent(otab, MS_TL_REPO_REF, tab_id, blob_id, auth_code, log_id, log_offset, q_time);
		
	exit_();
}

MSTempLogFile *MSDatabase::openTempLogFile(csWord4 log_id, size_t *log_rec_size, size_t *log_head_size)
{
	MSTempLog		*log;
	MSTempLogFile	*log_file = NULL;

	enter_();
	lock_(myTempLogArray);
	if (log_id)
		log = (MSTempLog *) myTempLogArray->get(log_id);
	else
		log = (MSTempLog *) myTempLogArray->first();
	if (log) {
		log_file = log->openTempLog();
		if (log_rec_size)
			*log_rec_size = log->myTemplogRecSize;
		if (log_head_size)
			*log_head_size = log->myTempLogHeadSize;
	}
	unlock_(myTempLogArray);
	return_(log_file);
}

u_int MSDatabase::getTempLogCount()
{
	u_int count;

	enter_();
	lock_(myTempLogArray);
	count = myTempLogArray->size();
	unlock_(myTempLogArray);
	return_(count);
}

void MSDatabase::removeTempLog(csWord4 log_id)
{
	enter_();
	lock_(myTempLogArray);
	myTempLogArray->remove(log_id);
	unlock_(myTempLogArray);
	exit_();
}

CSObject *MSDatabase::getKey()
{
	return (CSObject *) myDatabaseName;
}

int MSDatabase::compareKey(CSObject *key)
{
	return myDatabaseName->compare((CSString *) key);
}

MSCompactorThread *MSDatabase::getCompactorThread()
{
	return myCompactorThread;
}

CSSyncVector *MSDatabase::getRepositoryList()
{	
	return myRepostoryList;
}

csWord4 MSDatabase::registerBlobAlias(csWord4 repo_id, csWord8 repo_offset, const char *alias)
{
	csWord4 hash;
	bool can_retry = true;
	enter_();
	
retry:
	lock_(&iBlobAliaseLock);
	
	try_(a) {
		hash = iBlobAliases->addAlias(repo_id, repo_offset, alias);
	}
	
	catch_(a) {
		unlock_(&iBlobAliaseLock);
		if (can_retry) {
			// It can be that a duplicater alias exists that was deleted
			// but the transaction has not been written to the repository yet.
			// Flush all committed transactions to the repository file.
			MSTransactionManager::flush();
			can_retry = false;
			goto retry;
		}
		throw_();
	}
	
	cont_(a);
	unlock_(&iBlobAliaseLock);
	return_(hash);
}

csWord4 MSDatabase::updateBlobAlias(csWord4 repo_id, csWord8 repo_offset, csWord4 old_alias_hash, const char *alias)
{
	csWord4 new_hash;
	enter_();
	lock_(&iBlobAliaseLock);
	
	new_hash = iBlobAliases->addAlias(repo_id, repo_offset, alias);
	iBlobAliases->deleteAlias(repo_id, repo_offset, old_alias_hash);

	unlock_(&iBlobAliaseLock);
	return_(new_hash);
}

void MSDatabase::deleteBlobAlias(MSDiskAliasPtr diskRec)
{
	enter_();
	lock_(&iBlobAliaseLock);
	iBlobAliases->deleteAlias(diskRec);
	unlock_(&iBlobAliaseLock);
	exit_();
}

void MSDatabase::deleteBlobAlias(csWord4 repo_id, csWord8 repo_offset, csWord4 alias_hash)
{
	MSDiskAliasRec diskRec;
	
	CS_SET_DISK_4(diskRec.ar_repo_id_4, repo_id);	
	CS_SET_DISK_8(diskRec.ar_offset_8, repo_offset);	
	CS_SET_DISK_4(diskRec.ar_hash_4, alias_hash);
	deleteBlobAlias(&diskRec);
}

void MSDatabase::moveBlobAlias(csWord4 old_repo_id, csWord8 old_repo_offset, csWord4 alias_hash, csWord4 new_repo_id, csWord8 new_repo_offset)
{
	enter_();
	lock_(&iBlobAliaseLock);
	iBlobAliases->resetAlias(old_repo_id, old_repo_offset, alias_hash, new_repo_id, new_repo_offset);
	unlock_(&iBlobAliaseLock);
	exit_();
}

bool MSDatabase::isValidHeaderField(const char *name)
{
	bool is_valid = false;
	CSString		*header;
	enter_();

	if (name && *name) {
		if (strcasecmp(name, MS_ALIAS_TAG)) {
			lock_(&iHTTPMetaDataHeaders);
			header = CSString::newString(name);
			push_(header);
				
			is_valid = (iHTTPMetaDataHeaders.find(header) != NULL);
			release_(header);
			
			unlock_(&iHTTPMetaDataHeaders);
		} else 
			is_valid = true;
	}
	
	return_(is_valid);
}

void MSDatabase::setDefaultMetaDataHeaders(const char *defaults)
{
	enter_();
	
	if (!gDefaultMetaDataHeaders) 
		new_(gDefaultMetaDataHeaders, CSSyncStringBuffer());

	lock_(gDefaultMetaDataHeaders);
	if (defaults)
		gDefaultMetaDataHeaders->append(defaults);
	else
		gDefaultMetaDataHeaders->setLength(0);
	
	char *ptr = gDefaultMetaDataHeaders->getBuffer(0);
	for( u_int i = gDefaultMetaDataHeaders->length(); i; i--, ptr++) {
		if (*ptr == ':')
			*ptr = 0;
	}
	
	unlock_(gDefaultMetaDataHeaders);
	exit_();
}

#define METADATA_HEADER_FILE	"http-meta-data-headers"
void MSDatabase::loadMetaDataFields()
{
	CSPath			*path;
	CSString		*header;
	CSStringBuffer	*string;
	size_t			size = 0, len;
	char			*ptr;
	bool			save_it = false;


	enter_();
	myDatabasePath->retain();
	path = CSPath::newPath(myDatabasePath, METADATA_HEADER_FILE".dat");
	push_(path);
	if (!path->exists()) {
		CSPath *tmp_path;

		myDatabasePath->retain();
		tmp_path = CSPath::newPath(myDatabasePath, METADATA_HEADER_FILE".tmp");
		push_(tmp_path);
		if (tmp_path->exists())
			tmp_path->rename(METADATA_HEADER_FILE".dat");
		release_(tmp_path);
	}

	new_(string, CSStringBuffer(20));
	push_(string);

	if (path->exists()) {
		CSFile *file;


		file = path->openFile(CSFile::READONLY);
		push_(file);
		size = file->getEOF();
		string->setLength(size);
		file->read(string->getBuffer(0), 0, size, size);
		release_(file);

	} else {
		lock_(gDefaultMetaDataHeaders);
		size = gDefaultMetaDataHeaders->length() +1;
		string->setLength(size);
		strcpy(string->getBuffer(0), gDefaultMetaDataHeaders->getCString());	
		save_it = true;	
		unlock_(gDefaultMetaDataHeaders);
	}
	
	ptr = string->getBuffer(0);
	
	while (size) {
		header = CSString::newString(ptr);
		iHTTPMetaDataHeaders.add(header);
		
		len = strlen(ptr) +1;
		ptr += len;
		size -= len;
	}
		
	release_(string);

	release_(path);

	if (save_it)
		saveMetaDataFields();
	exit_();
}

void MSDatabase::saveMetaDataFields()
{
	CSPath			*path;
	CSPath			*old_path;
	CSFile			*file;
	CSCString		*header;
	size_t			offset = 0, len;
	char			null_char = 0;
	
	enter_();
	myDatabasePath->retain();
	path = CSPath::newPath(myDatabasePath, METADATA_HEADER_FILE".tmp");
	push_(path);

	file = path->openFile(CSFile::CREATE | CSFile::TRUNCATE);
	push_(file);
	u_int i = 0;
	while ((header = (CSCString *) iHTTPMetaDataHeaders.itemAt(i))) {
		len = header->length();
		file->write(header->getCString(), offset, len);
		offset += len;
		file->write(&null_char, offset, 1);
		offset++;
		i++;
	}
	file->close();
	release_(file);

	myDatabasePath->retain();
	old_path = CSPath::newPath(myDatabasePath, METADATA_HEADER_FILE".dat");
	push_(old_path);
	if (old_path->exists())
		old_path->remove();
	path->rename(METADATA_HEADER_FILE".dat");
	release_(old_path);

	release_(path);
	exit_();
}

void MSDatabase::startUp(const char *default_http_headers)
{
	enter_();
	
	new_(gDatabaseList, CSSyncSortedList);
	new_(gDatabaseArray, CSSparseArray(5));
	setDefaultMetaDataHeaders(default_http_headers);
	gBackupDatabaseID = 1;
	exit_();
}

void MSDatabase::stopThreads()
{
	MSDatabase *db;

	enter_();
	if (gDatabaseList) {
		lock_(gDatabaseList);
		for (int i=0;;i++) {
			if (!(db = (MSDatabase *) gDatabaseList->itemAt(i)))
				break;
			db->iClosing = true;
			
			if (db->myTempLogThread) {
				db->myTempLogThread->stop();
				db->myTempLogThread->release();
				db->myTempLogThread = NULL;
			}
			if (db->myCompactorThread) {
				db->myRepostoryList->wakeup(); // The compator thread waits on this.
				db->myCompactorThread->stop();
				db->myCompactorThread->release();
				db->myCompactorThread = NULL;
			}
			
			if (db->iBackupThread) {
				db->iBackupThread->stop();
				db->iBackupThread->release();
				db->iBackupThread = NULL;
			}
	
		}
		
		unlock_(gDatabaseList);
	}
	exit_();
}

void MSDatabase::shutDown()
{
		
	if (gDatabaseArray) {
		gDatabaseArray->clear();
		gDatabaseArray->release();
		gDatabaseArray = NULL;
	}
	
	if (gDatabaseList) {
		gDatabaseList->clear();
		gDatabaseList->release();
		gDatabaseList = NULL;
	}
	
	if (gDefaultMetaDataHeaders) {
		gDefaultMetaDataHeaders->release();
		gDefaultMetaDataHeaders = NULL;
	}
}

void MSDatabase::setBackupDatabase()
{
	enter_();
	// I need to give the backup database a unique fake database ID.
	// This is so that it is not confused with the database being backed
	// backed up when opening tables.
	
	// Normally database IDs are generated by time(NULL) so small database IDs
	// are safe to use as fake IDs.
	 
	lock_(gDatabaseList);
	myDatabaseID = gBackupDatabaseID++;
	retain();
	gDatabaseArray->set(myDatabaseID, this);
	isBackup = true;
	
	unlock_(gDatabaseList);
	
	// Rename the database path so that it is obviouse that this is an incomplete backup database.
	// When the backup is completed it will be renamed back.
	CSPath *new_path = CSPath::newPath(myDatabasePath->concat("#"));
	push_(new_path);
	
	if (new_path->exists()) 
		new_path->remove();
	
	myDatabasePath->retain();	
	CSPath *db_path = CSPath::newPath(myDatabasePath);
	push_(db_path);
	
	db_path->rename(new_path->getNameCString());
	myDatabasePath->release();
	myDatabasePath = new_path->getString();
	myDatabasePath->retain();
	
	release_(db_path);
	release_(new_path);
	
	
	exit_();
}

void MSDatabase::releaseBackupDatabase()
{
	enter_();

	
	// The backup has completed succefully, rename the path to the correct name.
	CSPath *db_path = CSPath::newPath(myDatabasePath->getCString());
	push_(db_path);
	
	myDatabasePath->setLength(myDatabasePath->length()-1);
	db_path->rename(cs_last_name_of_path(myDatabasePath->getCString()));
	release_(db_path);
	
	// Remove the backup database object.
	lock_(gDatabaseList);
	gDatabaseArray->remove(myDatabaseID);
	MSTableList::removeDatabaseTables(this); // Will also release the database object.
	unlock_(gDatabaseList);
	
	
	exit_();
}

void MSDatabase::startBackup()
{
	enter_();

	if (!myBackupLocation) {
		CSException::throwException(CS_CONTEXT, MS_ERR_INVALID_OPERATION, "A backup location has not been specified.");
	}
	
	if (iBackupThread) {
		if (iBackupThread->isRunning()) {
			CSException::throwException(CS_CONTEXT, MS_ERR_DUPLICATE_DB, "A backup is still running.");
		}
		iBackupThread->release();
		iBackupThread = NULL;
	}
	
	myDatabaseName->retain();
	myBackupLocation->retain();
	iBackupThread = MSBackup::newMSBackup(myBackupLocation, myDatabaseName, myDatabaseID);
	retain();
	
	iBackupThread->startBackup(this);
	
	exit_();
}

bool MSDatabase::backupStatus(csWord8 *total, csWord8 *completed, bool *completed_ok)
{
	bool done;
	
	enter_();
	
	if (iBackupThread) {
		*total = iBackupThread->getBackupSize();
		*completed = iBackupThread->getBackupCompletedSize();
		done = !iBackupThread->isRunning();
		*completed = (iBackupThread->getStatus() == 0);
	} else {
		*completed_ok = done = true;
		*total = *completed = 0;			
	}
		
	return_(done);
}

csWord4 MSDatabase::backupID()
{ 
	return (iBackupThread)?iBackupThread->backupID(): 0;
}

void MSDatabase::terminateBackup()
{
	if (iBackupThread) {
		iBackupThread->stop();
		iBackupThread->release();
		iBackupThread = NULL;
	}
}

MSDatabase *MSDatabase::getDatabase(CSString *db_name, bool create)
{
	MSDatabase *db;
	enter_();
	push_(db_name);
	
	
	lock_(gDatabaseList);
	if (!(db = (MSDatabase *) gDatabaseList->find(db_name))) {
		db_name->retain();
		db = MSDatabase::loadDatabase(db_name, create);
		if (!db)
			goto exit;
	} else
		db->retain();
	
	exit:
	unlock_(gDatabaseList);
	release_(db_name);
	return_(db);
}

MSDatabase *MSDatabase::getDatabase(const char *db_name, bool create)
{
	return getDatabase(CSString::newString(db_name), create);
}

MSDatabase *MSDatabase::getDatabase(csWord4 db_id)
{
	MSDatabase *db;
	
	enter_();
	lock_(gDatabaseList);
	if ((db = (MSDatabase *) gDatabaseArray->get((u_int) db_id))) 
		db->retain();
	else {
		// Look for the database folder with the correct ID:
		CSPath *path = CSPath::newPath(ms_my_get_mysql_home_path(), "pbms");
		push_(path);
		if (path->exists()) {
			CSDirectory *dir;
			path->retain();
			dir = CSDirectory::newDirectory(path);
			push_(dir);
			dir->open();
			
			while (dir->next() && !db) {
				if (!dir->isFile()) {
					const char *ptr, *dir_name  = dir->name();
					ptr = dir_name + strlen(dir_name) -1;
					
					while (ptr > dir_name && *ptr != '-') ptr--;
					
					if (*ptr ==  '-') {
						int len = ptr - dir_name;
						ptr++;
						if (atol(ptr) == db_id && len) {
							db = getDatabase(CSCString::newString(dir_name, len), true);
							ASSERT(db->myDatabaseID == db_id);
						}
					}
				}
			}
			release_(dir);
		}
		release_(path);		
	}
	unlock_(gDatabaseList);
	
	if (!db) {
		char buffer[CS_EXC_MESSAGE_SIZE];

		cs_strcpy(CS_EXC_MESSAGE_SIZE, buffer, "Unknown database #");
		cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, (u_int) db_id);
		CSException::throwException(CS_CONTEXT, MS_ERR_UNKNOWN_DB, buffer);
	}
	return_(db);
}

void MSDatabase::wakeTempLogThreads()
{
	MSDatabase *db;

	if (!gDatabaseList)
		return;
	enter_();
	lock_(gDatabaseList);
	for (int i=0;;i++) {
		if (!(db = (MSDatabase *) gDatabaseList->itemAt(i)))
			break;
		if (db->myTempLogThread)
			db->myTempLogThread->wakeup();
	}
	unlock_(gDatabaseList);
	exit_();
}

csWord4 MSDatabase::getDBID(CSPath *path, CSString *db_name, csWord4 requested_id)
{
	CSDirectory		*dir;
	csWord4			db_id = 0;
	int				len = db_name->length();
	const char		*ptr;
	
	enter_();
	push_(db_name);
	push_(path);
	
	// Search for the ID of the database
	path->retain();
	dir = CSDirectory::newDirectory(path);
	push_(dir);
	dir->open();
	while (dir->next() && !db_id) {
		if (!dir->isFile()){
			ptr = dir->name() + strlen(dir->name()) -1;
			while (ptr > dir->name() && isdigit(*ptr)) ptr--;
			if ((*ptr == '-') && (len == (ptr - dir->name())) && !db_name->compare(dir->name(), len) ) {
				db_id = atol(ptr+1);				
			}
		}
	}
	release_(dir);
	
	if (!db_id) {
		db_id = (requested_id)?requested_id: time(NULL);

		while (1) { // search for a unique db_id
			path->retain();
			dir = CSDirectory::newDirectory(path);
			push_(dir);
			dir->open();
			while (db_id && dir->next()) {
				if (!dir->isFile()) {
					ptr = dir->name() + strlen(dir->name()) -1;
					while (ptr > dir->name() && isdigit(*ptr)) ptr--;
					if ((*ptr == '-') && (db_id == atol(ptr+1))) {
						db_id = 0;				
					}
				}
			}
			release_(dir);
			if (db_id)
				break;
			sleep(1); // Allow 1 second to pass.
			db_id = time(NULL);
		} 
	}
	
	release_(path);
	release_(db_name);
	return_(db_id);
}

CSPath *MSDatabase::createDatabasePath(const char *location, CSString *db_name, csWord4 *db_id_ptr, bool *create)
{
	bool create_path = *create;	
	CSPath *path = NULL;
	char name_buffer[MS_DATABASE_NAME_SIZE + 40];
	csWord4 db_id;
	enter_();
	
	push_(db_name);
	*create = false;
	path = CSPath::newPath(location, "pbms");
	push_(path);
	if (!path->exists()) {
		if (!create_path){
			release_(path);
			path = NULL;
			goto done;
		}
			
		*create = true;
		path->makeDir();
	}

	path->retain();
	db_name->retain();
	db_id = getDBID(path, db_name, (db_id_ptr)?*db_id_ptr: 0);
	if (db_id_ptr)
		*db_id_ptr = db_id;
		
	// Create the PBMS database name with ID
	cs_strcpy(MS_DATABASE_NAME_SIZE + 40, name_buffer, db_name->getCString());
	cs_strcat(MS_DATABASE_NAME_SIZE + 40, name_buffer, "-");
	cs_strcat(MS_DATABASE_NAME_SIZE + 40, name_buffer, (u_int) db_id);
			
	pop_(path);
	path = CSPath::newPath(path, name_buffer);
	push_(path);
	if (!path->exists()) {
		if (create_path) {
			*create = true;
			path->makeDir();
		} else {
			release_(path);
			path = NULL;
		}
			
	}
	
done:
	if (path)
		pop_(path);
	release_(db_name);
	return_(path);
}



MSDatabase *MSDatabase::newDatabase(const char *db_location, CSString *db_name, csWord4	db_id, bool create)
{
	MSDatabase		*db = NULL;
	CSDirectory		*dir;
	MSRepository	*repo;
	CSPath			*path;
	const char		*file_name;
	u_int			file_id;
	off_t			file_size;
	MSTempLog		*log;
	u_int			to_delete = 0;
	CSString		*db_path;

	enter_();

	push_(db_name);

	/* Block the creation of the pbms database if there is no MySQL database. */
	db_name->retain();
	path = CSPath::newPath(ms_my_get_mysql_home_path(), db_name);
	push_(path);
	if (create && !path->exists()) {
		CSException::throwException(CS_CONTEXT, MS_ERR_UNKNOWN_DB, db_name->getCString());
	}
	release_(path);
	
	 // Create the database path, if 'create' == false then it can return NULL
	db_name->retain();
	path = createDatabasePath(db_location, db_name, &db_id, &create);
	if (!path) {
		release_(db_name);
		return_(NULL);
	}
	push_(path);
	
	// Create the database object and initialize it.
	if (!(db = new MSDatabase())) {
		CSException::throwOSError(CS_CONTEXT, ENOMEM);
	}
	
	db_path	= path->getString();
	db_path->retain();
	release_(path);
	path = NULL;
		
	db->iNextBlobRefId = (csWord4) time(NULL);
	db->iNextBlobRefId <<= 32;
	db->iNextBlobRefId = COMMIT_MASK(db->iNextBlobRefId);
	db->iNextBlobRefId++;
	
	db->myDatabaseID = db_id;
	db->myDatabasePath = db_path;
	db->myDatabaseName = db_name;
	db->myBlobCloud = NULL;

	pop_(db_name);	
	
	push_(db);
	
	
	new_(db->myTempLogArray, CSSyncSparseArray(20));
	new_(db->iTableList, CSSyncSortedList());
	new_(db->iTableArray, CSSparseArray(20));
	new_(db->myRepostoryList, CSSyncVector(20));
	//db->retain(); no retain here, MSAlias() takes a back ref.
	new_(db->iBlobAliases, MSAlias(db));
	
	/* "Load" the database: */

	/* Get the max table ID: */
	db_path->retain();
	dir = CSDirectory::newDirectory(db_path);
	push_(dir);
	dir->open();
	while (dir->next()) {
		file_name = dir->name();
		if (dir->isFile() && cs_is_extension(file_name, "bst"))
			db->addTableFromFile(dir, file_name, false);
	}
	release_(dir);

	db_path->retain();
	path = CSPath::newPath(db_path, "bs-repository");
	if (path->exists()) {
		dir = CSDirectory::newDirectory(path);
		push_(dir);
		dir->open();
		while (dir->next()) {
			file_name = dir->name();
			if (dir->isFile() && cs_is_extension(file_name, "bs")) {
				if ((file_id = ms_file_to_table_id(file_name, "repo"))) {
					dir->info(NULL, &file_size, NULL);
					new_(repo, MSRepository(file_id, db, file_size));
					db->myRepostoryList->set(file_id - 1, repo);
				}
			}
		}
		release_(dir);
	}
	else {
		path->makeDir();
		path->release();
	}

	db_path->retain();
	path = CSPath::newPath(db_path, "bs-logs");
	if (path->exists()) {
		dir = CSDirectory::newDirectory(path);
		push_(dir);
		dir->open();
		while (dir->next()) {
			file_name = dir->name();
			if (dir->isFile()) {
				if (cs_is_extension(file_name, "bs")) {
					if ((file_id = ms_file_to_table_id(file_name, "temp"))) {
						dir->info(NULL, &file_size, NULL);
						new_(log, MSTempLog(file_id, db, file_size));
						db->myTempLogArray->set(file_id, log);
					}
				}
				else if (cs_is_extension(file_name, "bst")) {
					db->addTableFromFile(dir, file_name, true);
					to_delete++;
				}
			}
		}
		release_(dir);
	}
	else {
		path->makeDir();
		path->release();
	}

	if (to_delete) {
		/* Go through and prepare all the tables that are to
		 * be deleted:
		 */
		u_int	i = 0;
		MSTable	*tab;
		
		while ((tab = (MSTable *) db->iTableList->itemAt(i))) {
			if (tab->isToDelete())
				tab->prepareToDelete();
			i++;
		}
	}


	pop_(db);

	return_(db);
}

void MSDatabase::startThreads()
{
	enter_();
	
	// iBlobAliases->ma_open() must be called before starting any threads.
	iBlobAliases->ma_open(); 

	new_(myTempLogThread, MSTempLogThread(1 * 1000, this));
	myTempLogThread->start();

#ifdef MS_COMPACTOR_POLLS
	new_(myCompactorThread, MSCompactorThread(MS_COMPACTOR_POLL_FREQ, this));
#else
	new_(myCompactorThread, MSCompactorThread(MS_DEFAULT_COMPACTOR_WAIT * 1000, this));
#endif

	
	myCompactorThread->start();
	exit_();
}


void MSDatabase::dropDatabase() 
{
	iDropping = true;
	iClosing = true;
	
	if (iBackupThread) {
		iBackupThread->stop();
		iBackupThread->release();
		iBackupThread = NULL;
	}
	
	if (myTempLogThread) {
		myTempLogThread->stop();
		myTempLogThread->release();
		myTempLogThread = NULL;
	}
	
	if (myCompactorThread) {
		myRepostoryList->wakeup(); // The compator thread waits on this.
		myCompactorThread->stop();
		myCompactorThread->release();
		myCompactorThread = NULL;
	}
	
	
}

void MSDatabase::removeDatabasePath(CSString *doomedDatabasePath ) 
{
	CSPath *path = NULL;
	CSDirectory *dir = NULL;
	const char *file_name;
	enter_();
	
	push_(doomedDatabasePath);
	
	// Delete repository files
	doomedDatabasePath->retain();
	path = CSPath::newPath(doomedDatabasePath, "bs-repository");
	push_(path);
	if (path->exists()) {
		path->retain();
		dir = CSDirectory::newDirectory(path);
		push_(dir);
		dir->open();
		while (dir->next()) {
			file_name = dir->name();
			if (dir->isFile() && cs_is_extension(file_name, "bs")) {
				dir->deleteEntry();
			}
		}
		release_(dir);
		if (path->isEmpty())
			path->removeDir();
	}
	release_(path);

	// Delete temp log files.
	doomedDatabasePath->retain();
	path = CSPath::newPath(doomedDatabasePath, "bs-logs");
	push_(path);
	if (path->exists()) {
		path->retain();
		dir = CSDirectory::newDirectory(path);
		push_(dir);
		dir->open();
		while (dir->next()) {
			file_name = dir->name();
			if (dir->isFile() && (cs_is_extension(file_name, "bs") || cs_is_extension(file_name, "bst"))) {
				dir->deleteEntry();
			}
		}
		release_(dir);
		if (path->isEmpty())
			path->removeDir();
	}
	release_(path);

	// Delete table reference files.
	doomedDatabasePath->retain();
	dir = CSDirectory::newDirectory(doomedDatabasePath);
	push_(dir);
	dir->open();
	while (dir->next()) {
		file_name = dir->name();
		if (dir->isFile() && cs_is_extension(file_name, "bst"))
			dir->deleteEntry();
	}
	release_(dir);

	doomedDatabasePath->retain();
	path = CSPath::newPath(doomedDatabasePath, "http-meta-data-headers.dat");
	push_(path);
	path->removeFile();
	release_(path);

	doomedDatabasePath->retain();
	path = CSPath::newPath(doomedDatabasePath, "ms_blob_alias.idx");
	push_(path);
	path->removeFile();
	release_(path);

	doomedDatabasePath->retain();
	path = CSPath::newPath(doomedDatabasePath);
	push_(path);
	if (path->isEmpty() && !path->isLink()) {
		path->removeDir();
	}
	release_(path);

	release_(doomedDatabasePath);
	
cleanup:
	path = CSPath::newPath(ms_my_get_mysql_home_path(), "pbms");
	push_(path);
	if (path->isEmpty() && !path->isLink()) {
		path->removeDir();
	}
	release_(path);

done:
	exit_();
}

/* Drop the PBMS database if it exists.
 * The root folder 'pbms' will be deleted also
 * if it is empty and not a symbolic link.
 * The database folder in 'pbms' is deleted if it is empty and
 * it is not a symbolic link.
 */
void MSDatabase::dropDatabase(MSDatabase *doomedDatabase, const char *db_name ) 
{
	CSString *doomedDatabasePath = NULL;

	enter_();
	
	if (doomedDatabase) {
		push_(doomedDatabase);
		doomedDatabasePath = doomedDatabase->myDatabasePath;
		
		doomedDatabasePath->retain();

		doomedDatabase->retain();
		MSTableList::removeDatabaseTables(doomedDatabase);

		doomedDatabase->dropDatabase(); // Shutdown database threads.
		
		// To avoid a deadlock a lock is not taken on the database list
		// if shutdown is in progress. The only database that would be
		// dropped during a shutdown is an incomplete backup database.
		ASSERT(doomedDatabase->isBackup || !self->myMustQuit);
		if (!self->myMustQuit) 
			lock_(gDatabaseList); // Be sure to shutdown the database before locking this or it can lead to deadlocks
			
		gDatabaseArray->remove(doomedDatabase->myDatabaseID);
		if (!doomedDatabase->isBackup)
			gDatabaseList->remove(doomedDatabase->getKey());
		if (!self->myMustQuit) 
			unlock_(gDatabaseList); 
		ASSERT(doomedDatabase->iRefCount == 1);
		release_(doomedDatabase);
		
	} else {
		CSPath *path;
		bool create = false;
		path = createDatabasePath(ms_my_get_mysql_home_path(), CSString::newString(db_name), NULL, &create);
		
		if (path) {
			push_(path);
			doomedDatabasePath = path->getString();
			doomedDatabasePath->retain();
			release_(path);
		}
	}
	
	if (doomedDatabasePath)
		removeDatabasePath(doomedDatabasePath);
	
done:
	exit_();
}

void MSDatabase::dropDatabase(const char *db_name ) 
{
	enter_();
	dropDatabase(getDatabase(db_name, false), db_name);
	exit_();
}

// The table_path can be several things here:
// 1: <absalute path>/<database>/<table>
// 2: <absalute path>/<database>
// 3: <database>/<table>
bool MSDatabase::convertTablePathToIDs(const char *table_path, csWord4 *db_id, csWord4 *tab_id, bool create) 
{
	const char	*base = ms_my_get_mysql_home_path();
	CSString	*table_url;
	CSString	*db_path = NULL;
	CSString	*db_name = NULL;
	CSString	*tab_name = NULL;
	MSDatabase	*db;
	enter_();
	
	*db_id = 0;
	*tab_id = 0;
	
	table_url = CSString::newString(table_path);
	if (table_url->startsWith(base)) {
		table_url = table_url->right(base);
	}
	push_(table_url);


	db_path = table_url->left("/", -1);
	push_(db_path);
	tab_name = table_url->right("/", -1);
	
	pop_(db_path);
	release_(table_url);

	if (db_path->length() == 0) { // Only a database name was supplied.
		db_path->release();
		db_name = tab_name;
		tab_name = NULL;
	} else {
		if (tab_name->length() == 0) {
			tab_name->release();
			tab_name = NULL;
		} else
			push_(tab_name);
		push_(db_path);
		db_name = db_path->right("/", -1);
		pop_(db_path);
		if (db_name->length() == 0) {
			db_name->release();
			db_name = db_path;
		} else {
			db_path->release();
			db_path = NULL;
		}
	}
	
	db = MSDatabase::getDatabase(db_name, create); // This will release db_name
	if (db) {
		*db_id = db->myDatabaseID;
		if (tab_name) {
			MSTable		*tab;
			pop_(tab_name);
			push_(db);
			tab = db->getTable(tab_name, create);// This will release tab_name
			pop_(db);
			if (tab) {
				*tab_id = tab->myTableID;
				tab->release();
			}
		}
			
		db->release();
	}
	
	return_((*tab_id > 0) && (*db_id > 0));
}

bool MSDatabase::convertTableAndDatabaseToIDs(const char *db_name, const char *tab_name, csWord4 *db_id, csWord4 *tab_id, bool create) 
{
	MSDatabase	*db;
	enter_();
	
	*db_id = 0;
	*tab_id = 0;
	
	db = MSDatabase::getDatabase(db_name, create); 
	if (db) {
		push_(db);
		*db_id = db->myDatabaseID;
		if (tab_name) {
			MSTable		*tab;
			tab = db->getTable(tab_name, create);
			if (tab) {
				*tab_id = tab->myTableID;
				tab->release();
			}
		}
			
		release_(db);
	}
	
	return_((*tab_id > 0) && (*db_id > 0));
}

MSDatabase *MSDatabase::loadDatabase(CSString *db_name, bool create)
{
	MSDatabase *db;
	enter_();
	
	db = newDatabase(ms_my_get_mysql_home_path(), db_name, 0, create);
	
	if (db) {
		push_(db);
		db->retain();
		gDatabaseList->add(db);
		db->retain();
		gDatabaseArray->set(db->myDatabaseID, db);
		db->startThreads();
		db->loadMetaDataFields();
	
		pop_(db);
	}
	return_(db);
}

csWord4 MSDatabase::getDatabaseID(CSString *db_name, bool create)
{
	MSDatabase *db;
	csWord4 id = 0;
	enter_();
	push_(db_name);
	
	
	lock_(gDatabaseList);
	if (!(db = (MSDatabase *) gDatabaseList->find(db_name))) {
		db_name->retain();
		db = MSDatabase::loadDatabase(db_name, create);
		if (!db)
			goto exit;
		id = db->myDatabaseID;
		db->release();
	} else
		id = db->myDatabaseID;
	
	exit:
	unlock_(gDatabaseList);
	release_(db_name);
	return_(id);
}


csWord4 MSDatabase::getDatabaseID(const char *db_name, bool create)
{
	return getDatabaseID(CSString::newString(db_name), create);
}

MSDatabase *MSDatabase::getBackupDatabase(CSString *db_location, CSString *db_name, csWord4 db_id, bool create)
{
	bool was_created = create; 
	csWord4 create_db_id = db_id;
	CSPath *path;
	MSDatabase *db;
	enter_();
	
	push_(db_location);
	push_(db_name);
	// If the db already exists and 'create' == true then the existing db
	// must be deleted.
	// Create the database path, if 'create' == false then it can return NULL
	db_name->retain();
	path = createDatabasePath(db_location->getCString(), db_name, &create_db_id, &was_created);
	if (!path) {
		CSException::throwException(CS_CONTEXT, MS_ERR_UNKNOWN_DB, db_name->getCString());
	}
	push_(path);
	
	// If we wanted to create it but it already exists then throw an error.
	if ( create && !was_created) {
		CSException::throwException(CS_CONTEXT, MS_ERR_DUPLICATE_DB, db_name->getCString());
	}
	
	// Check that the database ID we want to use was available.
	if (db_id != create_db_id) {
		if (was_created) {
			CSString *doomedDatabasePath = path->getString();
			doomedDatabasePath->retain();
			removeDatabasePath(doomedDatabasePath);
		}
		CSException::throwException(CS_CONTEXT, MS_ERR_DUPLICATE_DB_ID, db_name->getCString());
	}
	
	release_(path);
	pop_(db_name);
	// everything looks OK
	db = newDatabase(db_location->getCString(), db_name, db_id, create);
	db->setBackupDatabase();
	release_(db_location);
	return_(db);
}

