/* Copyright (c) 2009 PrimeBase Technologies GmbH, Germany
 *
 * PrimeBase Media Stream for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Barry Leslie
 *
 * System variables table.
 *
 */
#include "CSConfig.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <time.h>

//#include "mysql_priv.h"
#include "CSGlobal.h"
#include "CSStrUtil.h"

#include "ha_pbms.h"
//#include <plugin.h>

#include "ms_mysql.h"
#include "Repository_ms.h"
#include "Database_ms.h"
#include "Compactor_ms.h"
#include "OpenTable_ms.h"
#include "Util_ms.h"
#include "Discover_ms.h"


#include "SysTab_variable.h"

#define MS_REPOSITORY_STORAGE_TYPE	"REPOSITORY"
#define MS_CLOUD_STORAGE_TYPE		"CLOUD"

DT_FIELD_INFO pbms_variable_info[]=
{
	{"Id",			NULL,	NULL, MYSQL_TYPE_LONG,		NULL,							NOT_NULL_FLAG,	"The variable ID"},
	{"Name",		32,		NULL, MYSQL_TYPE_VARCHAR,	&my_charset_utf8_general_ci,	NOT_NULL_FLAG,	"PBMS variable name"},
	{"Value",	1024,	NULL, MYSQL_TYPE_VARCHAR,	&my_charset_utf8_general_ci,	0,	"PBMS variable value."},
	{"Description",	124,	NULL, MYSQL_TYPE_VARCHAR,	&my_charset_utf8_general_ci,	NOT_NULL_FLAG,	"PBMS variable description."},
	{NULL,NULL, NULL, MYSQL_TYPE_STRING,NULL, 0, NULL}
};

DT_KEY_INFO pbms_variable_keys[]=
{
	{"pbms_variable_pk", PRI_KEY_FLAG, {"Id", NULL}},
	{NULL, 0, {NULL}}
};

typedef const char *(*PBMSVarGetFunc)(MSDatabase *db, const char *dflt_value);
typedef const char *(*PBMSVarCheckFunc)(char *value, bool *ok);
typedef void (*PBMSVarActionFunc)(MSDatabase *db, const char *value);

typedef struct {
	const char *name;
	const char *value;
	const char *info;
	bool save;			// true if the value is to be saved to disk on update
	PBMSVarGetFunc get;
	PBMSVarCheckFunc check;
	PBMSVarActionFunc action;
} PBMSVariableRec, *PBMSVariablePtr;

//---------------------------
static char *cleanupVariable(char *value, int *len)
{
	char *ptr;
	
	while (*value && isspace(*value)) value++;
	ptr = value + strlen(value) -1;
	while ((ptr > value) && isspace(*ptr)) ptr--;
	ptr++;
	*ptr = 0;
	
	ptr = value;
	while (*ptr) {
		*ptr = toupper(*ptr);
		ptr++;
	}
	
	*len = ptr - value;
	return value;
}

//---------------------------
static const char *get_dflt(MSDatabase *db, const char *dflt_value)
{
	db->release();
	return dflt_value;
}

//---------------------------
static const char *get_restore_dump(MSDatabase *db, const char *dflt_value)
{
	const char *value;
	enter_();
	push_(db);
	if (db->isRecovering())
		value = "TRUE";
	else
		value = "FALSE";
	release_(db);
	return_(value);
}

//---------------------------
static const char *boolCheck(char *value, bool *ok)
{
	const char *val = "Invalid boolean variable, try 'true' or 'false'";
	int len;
	
	value = cleanupVariable(value, &len);
	
	*ok = false;
	switch (*value) {
		case '0':
			if (len == 1) {
				*ok = true;
				val = "FALSE";
			}
			break;
			
		case '1':
			if (len == 1) {
				*ok = true;
				val = "TRUE";
			}
			break;
			
		case 'T':
			if (!strcmp(value, "TRUE")) {
				*ok = true;
				val = "TRUE";
			}
			break;
			
		case 'F':
			if (!strcmp(value, "FALSE")) {
				*ok = true;
				val = "FALSE";
			}
			break;
			
				
	}
	
	return val;
}

//---------------------------
static const char *storageTypeCheck(char *value, bool *ok)
{
	const char *val = "Invalid storage type, try '"MS_REPOSITORY_STORAGE_TYPE"' or '"MS_CLOUD_STORAGE_TYPE"'";
	int len;
	
	value = cleanupVariable(value, &len);
	*ok = false;
	
	if (!strcmp(value, MS_REPOSITORY_STORAGE_TYPE)) {
		*ok = true;
		val = MS_REPOSITORY_STORAGE_TYPE;
	} else if (!strcmp(value, MS_CLOUD_STORAGE_TYPE)) {
		*ok = true;
		val = MS_CLOUD_STORAGE_TYPE;
	}
	
	return val;
}

//---------------------------
static void do_DumpRestore(MSDatabase *db, const char *value)
{
	enter_();
	push_(db);
	db->setRecovering(strcmp(value, "FALSE"));
	release_(db);
	exit_();
}

//---------------------------
static void do_SetStorageType(MSDatabase *db, const char *value)
{
	enter_();
	push_(db);
	 // TO DO:
	release_(db);
	exit_();
}

//---------------------------
static void do_SetS3PublicKey(MSDatabase *db, const char *value)
{
	enter_();
	push_(db);
	 // TO DO:
	release_(db);
	exit_();
}

//---------------------------
static void do_SetS3PrivateKey(MSDatabase *db, const char *value)
{
	enter_();
	push_(db);
	 // TO DO:
	release_(db);
	exit_();
}

//---------------------------
static void do_SetS3Bucket(MSDatabase *db, const char *value)
{
	enter_();
	push_(db);
	 // TO DO:
	release_(db);
	exit_();
}

//---------------------------
static const char *get_BackupLocation(MSDatabase *db, const char *dflt_value)
{
	const char *value;
	
	value = db->getBackupLocation();
	db->release();
	return value;
}

//---------------------------
static void do_setBackupLocation(MSDatabase *db, const char *value)
{
	enter_();
	push_(db);
	db->setBackupLocation(value);
	release_(db);
	exit_();
}

//---------------------------
static const char *get_Backup(MSDatabase *db, const char *dflt_value)
{
	const char *value;
	csWord8 total, completed;
	bool completed_ok;
	
	if ( db->backupStatus(&total, &completed, &completed_ok))
		value = "FALSE";
	else
		value = "TRUE";
	
	db->release();
	return value;
}
//---------------------------
static void do_Backup(MSDatabase *db, const char *value)
{
	enter_();
	push_(db);
	if (!strcmp(value, "TRUE"))
		db->startBackup();
	else
		db->terminateBackup();
	release_(db);
	exit_();
}

//---------------------------
static void do_Restore(MSDatabase *db, const char *value)
{
	enter_();
	push_(db);
	if (!strcmp(value, "TRUE"))
		;//db->restoreRepository();
	release_(db);
	exit_();
}


static PBMSVariableRec variables[] = {
	{"Storage-type", MS_REPOSITORY_STORAGE_TYPE, "How the BLOB data is to be stored.", true, get_dflt, storageTypeCheck, do_SetStorageType},
	{"S3-Public-Key", NULL, "The public key for S3 storage.", true, get_dflt, NULL, do_SetS3PublicKey},
	{"S3-Private-Key", NULL, "The private key for S3 storage.", true, get_dflt, NULL, do_SetS3PrivateKey},
	{"S3-Bucket-Name", NULL, "The S3 bucket to be used for S3 storage.", true, get_dflt, NULL, do_SetS3Bucket},
	{"Restoring-Dump", "FALSE", "Indicate if the database is being restored from a dump file.", false, get_restore_dump, boolCheck, do_DumpRestore},
	{"Backup-Location", NULL, "The location for repository level backup and restore operations.", true, get_BackupLocation, NULL, do_setBackupLocation},
	{"Do-Backup", "FALSE", "When true the repository will be backed up to the backup location.", false, get_Backup, boolCheck, do_Backup}
//	{"Do-Restore", "FALSE", "When true the repository will be restored from the backup location.", false, get_dflt, boolCheck, do_Restore}
};

static csWord4 num_variables = 7;

MSVariableTable::MSVariableTable(MSSystemTableShare *share, TABLE *table):
MSOpenSystemTable(share, table),
iVariableIndex(0)
{
}

MSVariableTable::~MSVariableTable()
{
	//unuse();
}

void MSVariableTable::use()
{
	myShare->mySysDatabase->iHTTPMetaDataHeaders.lock();
}

void MSVariableTable::unuse()
{
	myShare->mySysDatabase->iHTTPMetaDataHeaders.unlock();
	
}


void MSVariableTable::seqScanInit()
{
	iVariableIndex = 0;
}

bool MSVariableTable::seqScanNext(char *buf)
{
	TABLE		*table = mySQLTable;
	Field		*curr_field;
	byte		*save;
	MY_BITMAP	*save_write_set;
	PBMSVariablePtr var;
	
	enter_();
	
	if (iVariableIndex >= num_variables)
		return_(false);
	var = &(variables[iVariableIndex++]);
	
	save_write_set = table->write_set;
	table->write_set = NULL;

	memset(buf, 0xFF, table->s->null_bytes);
 	for (Field **field=table->field ; *field ; field++) {
 		curr_field = *field;
		save = curr_field->ptr;
#if MYSQL_VERSION_ID < 50114
		curr_field->ptr = (byte *) buf + curr_field->offset();
#else
		curr_field->ptr = (byte *) buf + curr_field->offset(curr_field->table->record[0]);
#endif
		switch (curr_field->field_name[0]) {
			case 'I':
				ASSERT(strcmp(curr_field->field_name, "Id") == 0);
				curr_field->store(iVariableIndex, true);
				break;

			case 'N':
				ASSERT(strcmp(curr_field->field_name, "Name") == 0);
					curr_field->store(var->name, strlen(var->name), &my_charset_utf8_general_ci);
					ms_my_set_notnull_in_record(curr_field, buf);
				break;

			case 'V': {
				ASSERT(strcmp(curr_field->field_name, "Value") == 0);
					const char *value;
					myShare->mySysDatabase->retain();
					value = var->get(myShare->mySysDatabase, var->value);
					if (value) {
						curr_field->store(value, strlen(value), &my_charset_utf8_general_ci);
						ms_my_set_notnull_in_record(curr_field, buf);
					}
				}
				break;

			case 'D':
				ASSERT(strcmp(curr_field->field_name, "Description") == 0);
					curr_field->store(var->info, strlen(var->info), &my_charset_utf8_general_ci);
					ms_my_set_notnull_in_record(curr_field, buf);
				break;

		}
		curr_field->ptr = save;
	}

	table->write_set = save_write_set;
	return_(true);
}

void MSVariableTable::seqScanPos(csWord1 *pos)
{
	csInt4 index = iVariableIndex -1;
	if (index < 0)
		index = 0; // This is probably an error condition.
		
	mi_int4store(pos, index);
}

void MSVariableTable::seqScanRead(csWord1 *pos, char *buf)
{
	iVariableIndex = mi_uint4korr(pos);
	seqScanNext(buf);
}

void MSVariableTable::updateRow(char *old_data, char *new_data) 
{
	TABLE	*table = mySQLTable;	
	csWord4 n_id, o_id;
	String n_str1, n_str2, *n_var_name = &n_str1, *n_var_value = &n_str2;	
	String o_str1, *o_var_name = &o_str1;	
	const char *clean_value;

	enter_();
	
	GET_INT_FIELD(o_id, 0, table, old_data);
	GET_STR_FIELD(o_var_name, 1, table, old_data);

	GET_INT_FIELD(n_id, 0, table, new_data);
	GET_STR_FIELD(n_var_name, 1, table, new_data);
	GET_STR_FIELD(n_var_value, 2, table, new_data);

	// The command names must match.
	if ((n_id != o_id) || my_strcasecmp(&my_charset_utf8_general_ci, o_var_name->c_ptr_safe(), n_var_name->c_ptr_safe()))
		CSException::throwException(CS_CONTEXT, HA_ERR_TABLE_READONLY, "Attempt to update read only fields in the PBMS command table.");
		
	n_id--;
	if (n_id >  num_variables) // Should never happen
		CSException::throwException(CS_CONTEXT, HA_ERR_KEY_NOT_FOUND, "Invalid id");
	
	CSStringBuffer *value;	
	new_(value, CSStringBuffer(0));
	push_(value);
	value->append(n_var_value->c_ptr(), n_var_value->length());
	
	// check the input value converting it to a standard format where aplicable:
	if (variables[n_id].check) {
		bool ok = false;
		clean_value = variables[n_id].check(value->getCString(), &ok);
		if (!ok)
			CSException::throwException(CS_CONTEXT, HA_ERR_GENERIC, clean_value);
	} else
		clean_value = value->getCString();
		
	// Execute the action associated with the variable.
	if (variables[n_id].action) {
		myShare->mySysDatabase->retain();
		variables[n_id].action(myShare->mySysDatabase, clean_value);
	}
	
	release_(value);
	
	exit_();
}

